import os
import platform
from pathlib import Path
from typing import Optional

__os_folder_names = {"Linux": "linux_x64", "Darwin": "osX", "Windows": "win64"}

try:
    os_system_name = platform.system()
    os_folder_name = __os_folder_names[os_system_name]
except KeyError:
    raise ValueError(
        f"Platform {platform.system()} is not recognised as a valid platform. "
        f"Valid platforms are : {__os_folder_names.keys()}"
    )

_IS_WINDOWS_OS = os_system_name == "Windows"

packaged_binaries_folder = (
    Path(__file__).parent.parent.parent / "bin" / "nifty_reg"
)


def conda_niftyreg_path() -> Optional[Path]:
    """
    If a conda install of niftyreg is available, return the directory
    containing the niftyreg binaries.
    """
    if "CONDA_PREFIX" in os.environ:
        conda_prefix = Path(os.environ["CONDA_PREFIX"])

        if platform.system() == "Windows":
            # Install prefix into conda environments on Windows
            # is CONDA_PREFIX / Library / bin
            bin_path = conda_prefix / "Library" / "bin"

            # Determine if the binaries are present
            # Binaries MUST have .exe appended to them in this case
            if (bin_path / "reg_aladin.exe").exists():
                return bin_path
        else:
            # On MacOS and Linux, binaries are placed into
            # CONDA_PREFIX / bin
            bin_path = conda_prefix / "bin"

            # Determine if the binaries are present
            if (bin_path / "reg_aladin").exists():
                return bin_path
    # If the binaries are not installed into the conda environment,
    # or there is no conda environment, return None as a fail-case
    return None


_CONDA_NIFTYREG_BINARY_PATH = conda_niftyreg_path()


def get_binary(program_name: str) -> Path:
    """
    Get path to one of the niftyreg binaries.

    If niftyreg is installed via conda, use those binaries, otherwise fall
    back on bundled binaries.
    """
    if _CONDA_NIFTYREG_BINARY_PATH is not None:
        bin_path = _CONDA_NIFTYREG_BINARY_PATH / program_name
    else:
        bin_path = packaged_binaries_folder / os_folder_name / program_name

    # Append exe label to Windows executables
    # It looks like subprocess is actually able to cope without the .exe
    # appended, but just to be safe we'll include it on Windows OS calls
    if _IS_WINDOWS_OS:
        bin_path = bin_path.parent / f"{bin_path.stem}.exe"
    return bin_path



import logging

import bg_space as bg
import imio
from bg_atlasapi import BrainGlobeAtlas
from brainglobe_utils.general.system import get_num_processes

from brainreg.core.backend.niftyreg.run import run_niftyreg
from brainreg.core.exceptions import (
    LoadFileException,
    path_is_folder_with_one_tiff,
)
from brainreg.core.utils.boundaries import boundaries
from brainreg.core.utils.volume import calculate_volumes


def main(
    atlas,
    data_orientation,
    target_brain_path,
    paths,
    voxel_sizes,
    niftyreg_args,
    preprocessing_args,
    n_free_cpus=2,
    sort_input_file=False,
    additional_images_downsample=None,
    backend="niftyreg",
    scaling_rounding_decimals=5,
    debug=False,
    save_original_orientation=False,
    brain_geometry="full",
):
    if path_is_folder_with_one_tiff(target_brain_path):
        raise LoadFileException("single_tiff")

    atlas = BrainGlobeAtlas(atlas)
    source_space = bg.AnatomicalSpace(data_orientation)

    scaling = []
    for idx, axis in enumerate(atlas.space.axes_order):
        scaling.append(
            round(
                float(voxel_sizes[idx])
                / atlas.resolution[
                    atlas.space.axes_order.index(source_space.axes_order[idx])
                ],
                scaling_rounding_decimals,
            )
        )

    n_processes = get_num_processes(min_free_cpu_cores=n_free_cpus)
    load_parallel = n_processes > 1

    logging.info("Loading raw image data")

    try:
        target_brain = imio.load_any(
            target_brain_path,
            scaling[1],
            scaling[2],
            scaling[0],
            load_parallel=load_parallel,
            sort_input_file=sort_input_file,
            n_free_cpus=n_free_cpus,
        )
    except ValueError as error:
        raise LoadFileException(None, error) from None

    target_brain = bg.map_stack_to(
        data_orientation, atlas.metadata["orientation"], target_brain
    )

    if backend == "niftyreg":
        run_niftyreg(
            paths.registration_output_folder,
            paths,
            atlas,
            target_brain,
            n_processes,
            additional_images_downsample,
            data_orientation,
            atlas.metadata["orientation"],
            niftyreg_args,
            preprocessing_args,
            scaling,
            load_parallel,
            sort_input_file,
            n_free_cpus,
            debug=debug,
            save_original_orientation=save_original_orientation,
            brain_geometry=brain_geometry,
        )

    logging.info("Calculating volumes of each brain area")
    calculate_volumes(
        atlas,
        paths.registered_atlas,
        paths.registered_hemispheres,
        paths.volume_csv_path,
        # for all brainglobe atlases
        left_hemisphere_value=1,
        right_hemisphere_value=2,
        brain_geometry=brain_geometry,
    )

    logging.info("Generating boundary image")
    boundaries(paths.registered_atlas, paths.boundaries_file_path)

    logging.info(
        f"brainreg completed. Results can be found here: "
        f"{paths.registration_output_folder}"
    )



from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("brainreg")
    del version
except PackageNotFoundError:
    # package is not installed
    pass

__author__ = "Adam Tyson"



import os
import traceback as tb
from os.path import isdir


def path_is_folder_with_one_tiff(target_brain_path):
    """
    Check that the target dir contains only 1 TIFF.
    In this case, the path to the .tiff should be
    passed not the directory.
    """
    if isdir(target_brain_path):
        image_files = search_dir_for_image_files(target_brain_path)

        if len(image_files) == 1:
            return True


def search_dir_for_image_files(target_brain_path):
    """
    Search a folder for any images that can be loaded by brainreg

    """
    images_in_dir = [
        file
        for file in os.listdir(target_brain_path)
        if file.lower().endswith((".tif", ".tiff"))
    ]
    return images_in_dir


class LoadFileException(Exception):
    """
    Custom exception class for errors found loading image with
    imio.load_any (in main.py).

    If the passed target brain directory contains only a single
    .tiff, alert the user.
    Otherwise, alert the use there was an issue loading the file and
    including the full traceback

    Set the error message to self.message to read during testing.
    """

    def __init__(self, error_type=None, base_error=None):
        if error_type == "single_tiff":
            self.message = (
                "Attempted to load directory containing a "
                "single .tiff file. If the .tiff file is 3D "
                "please pass the full path with filename. "
                "Single 2D .tiff file input is "
                "not supported."
            )

        else:
            origional_traceback = "".join(
                tb.format_tb(base_error.__traceback__) + [base_error.__str__()]
            )
            self.message = (
                f"{origional_traceback}\nFile failed to load with "
                f"imio. Ensure all image files contain the "
                f"same number of pixels. Full traceback above."
            )

        super().__init__(self.message)
